use futures::stream::StreamExt;
use std::{io, str::FromStr};

use irc::{
    client::{Client, data::Config},
    proto::Command,
};

type Error = Box<dyn std::error::Error + Send + Sync>;
type Result<T> = std::result::Result<T, Error>;

#[derive(Debug)]
pub struct CapCode {
    pub capcode: &'static str,
    pub region: &'static str,
    pub location: &'static str,
    pub service: &'static str,
    pub description: &'static str,
}

include!(concat!(env!("OUT_DIR"), "/capcodes.rs"));

#[allow(unused)]
#[derive(Debug)]
struct Message {
    raw: String,
    protocol: String,
    timestamp: String,
    flags: String,
    frameid: String,
    capcodes_raw: String,
    capcodes: Vec<&'static CapCode>,
    format: String,
    message: String,
}

impl std::fmt::Display for Message {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        if self.capcodes.is_empty() {
            write!(f, "[{}]: {}", self.capcodes_raw, self.message)?;
            return Ok(());
        }

        let mut iterator = self.capcodes.iter().peekable();
        while let Some(code) = iterator.next() {
            write!(f, "[{}]", code.service)?;
            if !code.region.is_empty() {
                write!(f, "[{}]", code.region.trim())?;
            }
            if !code.location.is_empty() {
                write!(f, "[{}]", code.location.trim())?;
            }
            if !code.description.is_empty() {
                write!(f, "[{}]", code.description.trim())?;
            }
            write!(f, ": ")?;
            write!(f, "{}", self.message)?;
            if iterator.peek().is_some() {
                writeln!(f)?;
            }
        }
        Ok(())
    }
}

impl FromStr for Message {
    type Err = Error;

    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
        let mut parts = s.split('|'); //.collect::<Vec<&str>>();
        if parts.clone().count() != 7 {
            return Err("Invalid message".into());
        }

        let Some(protocol) = parts.next() else {
            return Err("Missing Protocol".into());
        };
        let Some(timestamp) = parts.next() else {
            return Err("Missing Timestamp".into());
        };
        let Some(flags) = parts.next() else {
            return Err("Missing Flags".into());
        };
        let Some(frameid) = parts.next() else {
            return Err("Missing FrameID".into());
        };
        let Some(capcodes) = parts.next() else {
            return Err("Missing Cap Codes".into());
        };
        let Some(format) = parts.next() else {
            return Err("Missing Format".into());
        };
        let Some(message) = parts.next() else {
            return Err("Missing Message".into());
        };

        let mut cc = vec![];
        let parts = capcodes.split(' ');
        for c in parts {
            let c = &c[2..];
            let Some(cap) = CAPCODES.get(c) else {
                continue;
            };
            cc.push(cap);
        }

        Ok(Self {
            raw: s.to_string(),
            protocol: protocol.to_string(),
            timestamp: timestamp.to_string(),
            flags: flags.to_string(),
            frameid: frameid.to_string(),
            capcodes_raw: capcodes.to_string(),
            capcodes: cc,
            format: format.to_string(),
            message: message.trim().to_string(),
        })
    }
}

async fn real_main() -> Result<()> {
    dotenvy::dotenv().ok();

    let irc_server = std::env::var("IRC_SERVER")?;
    let irc_port: u16 = std::env::var("IRC_PORT")?.parse()?;
    let irc_tls: bool = std::env::var("IRC_TLS")?.parse()?;
    let irc_nick = std::env::var("IRC_NICK")?;
    let irc_password = std::env::var("IRC_PASSWORD")?;

    let mut buffer = String::new();
    let stdin = io::stdin(); // We get `Stdin` here.

    let (term_tx, term_rx) = tokio::sync::broadcast::channel::<()>(1);
    let (tx, mut rx) = tokio::sync::mpsc::channel(12);

    let config = Config {
        nickname: Some(irc_nick),
        server: Some(irc_server),
        port: Some(irc_port),
        use_tls: Some(irc_tls),
        channels: vec!["#p2000".to_owned()],
        ..Config::default()
    };
    let mut client = Client::from_config(config).await?;
    client.identify()?;

    let mut stream = client.stream()?;
    let mut term_rx_a = term_rx.resubscribe();
    tokio::spawn(async move {
        loop {
            tokio::select! {
                Some(msg) = rx.recv() => {
                    client.send(msg)?;
                },
                _ = term_rx_a.recv() => {
                    eprintln!("Closing irc send loop");
                    break;
                }
            }
        }

        Ok::<(), Error>(())
    });

    let tx_ = tx.clone();
    let mut term_rx_a = term_rx.resubscribe();
    tokio::spawn(async move {
        loop {
            tokio::select! {
                message = stream.next() => {
                    let Ok(Some(message)) = message.transpose() else {
                        return Err("".into());
                    };
                    print!("{}", message);

                    if let Command::NOTICE(a, b) = message.command {
                        if a == "p2000" && b.starts_with("This nickname is registered.") {
                            tx_.send(Command::NICKSERV(vec![
                                "identify".into(),
                                irc_password.clone(),
                            ]))
                            .await?;
                        }
                    }
                },
                _ = term_rx_a.recv() => {
                    eprintln!("Closing irc receiver loop");
                    break;
                }
            }
        }

        Ok::<(), Error>(())
    });

    'outer: while stdin.read_line(&mut buffer)? > 0 {
        if let Ok(msg) = Message::from_str(&buffer) {
            for msg in msg.to_string().lines() {
                println!("{}", msg);
                if let Err(e) = tx
                    .send(Command::PRIVMSG("#p2000".to_string(), msg.to_string()))
                    .await
                {
                    // shut everything down, we need to reconnect/restart
                    eprintln!("{:?}", e);
                    let _ = term_tx.send(());
                    break 'outer;
                };
            }
        }
        buffer.clear();
    }

    eprintln!("End of program...");

    Ok(())
}

#[tokio::main]
async fn main() -> Result<()> {
    loop {
        if let Err(e) = real_main().await {
            eprintln!("{:?}", e);
            continue;
        }
        eprintln!("Restarting...");
    }
}
